/*
 *   Copyright 2017-2020, Schlumberger
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package org.opengroup.osdu.notification.api;

import com.fasterxml.jackson.databind.ObjectMapper;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.opengroup.osdu.core.common.cryptographic.ISignatureService;
import org.opengroup.osdu.core.common.http.HttpClient;
import org.opengroup.osdu.core.common.http.HttpResponse;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.core.common.model.notification.Subscription;
import org.opengroup.osdu.core.common.notification.ISubscriptionService;
import org.opengroup.osdu.core.common.notification.SubscriptionException;
import org.opengroup.osdu.core.common.notification.SubscriptionFactory;
import org.opengroup.osdu.core.common.notification.SubscriptionService;
import org.opengroup.osdu.notification.di.CredentialHeadersProvider;
import org.opengroup.osdu.notification.di.SubscriptionCacheFactory;
import org.opengroup.osdu.notification.provider.interfaces.IGoogleServiceAccount;
import org.opengroup.osdu.notification.provider.interfaces.IPubsubRequestBodyExtractor;
import org.powermock.modules.junit4.PowerMockRunner;
import org.springframework.http.ResponseEntity;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import static org.junit.Assert.fail;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

@RunWith(PowerMockRunner.class)
public class PubsubEndpointTests {
    @Mock
    private IPubsubRequestBodyExtractor pubsubRequestBodyExtractor;
    @Mock
    private ISignatureService signatureService;
    @Mock
    private HttpClient httpClient;
    @Mock
    private DpsHeaders headers;
    @Mock
    private SubscriptionCacheFactory subscriptionCacheFactory;
    @Mock
    private IGoogleServiceAccount googleIdTokenProducer;
    @Mock
    private JaxRsDpsLog log;
    @InjectMocks
    private PubsubEndpoint sut;
    @Mock
    private CredentialHeadersProvider credentialHeadersProvider;
    @Mock
    private SubscriptionFactory subscriptionFactory;
    @Mock
    private ObjectMapper objectMapper;

    private static final String SIGNED_SIGNATURE = "testEncodedInfo.testSignature";
    private static final String GOOGLE_ID_TOKEN = "testHeader.testPayload.testSignature";
    private static final String NOTIFICATION_ID = "test-notification-id";
    private static final String PUBSUB_MESSAGE = "test-pubsub-message-data";

    private HttpResponse response = new HttpResponse();

    @Before
    public void setup() throws Exception {
        when(this.pubsubRequestBodyExtractor.extractNotificationIdFromRequestBody()).thenReturn(NOTIFICATION_ID);
        when(this.pubsubRequestBodyExtractor.extractDataFromRequestBody()).thenReturn(PUBSUB_MESSAGE);
    }

    @Test
    public void should_return200_whenPubsubMessageValidAndSuccessCodeReturnedFromClient_hmac() throws Exception {
        response.setResponseCode(200);

        when(this.signatureService.getSignedSignature(any(), any())).thenReturn(SIGNED_SIGNATURE);
        when(this.httpClient.send(any())).thenReturn(response);
        when(this.subscriptionCacheFactory.get(any())).thenReturn(getHmacSubscription());
        sut.setObjectMapper(new ObjectMapper());
        ResponseEntity responseEntity = this.sut.recordChanged();
        Assert.assertEquals(200, responseEntity.getStatusCode().value());
        Assert.assertEquals("message acknowledged by client", responseEntity.getBody().toString());
    }

    @Test
    public void should_return200_whenPubsubMessageValidAndSuccessCodeReturnedFromClient_gsa() throws Exception {
        response.setResponseCode(200);

        when(this.googleIdTokenProducer.getIdToken(any(), any())).thenReturn(GOOGLE_ID_TOKEN);
        when(this.httpClient.send(any())).thenReturn(response);
        when(this.subscriptionCacheFactory.get(any())).thenReturn(getGsaSubscription());
        sut.setObjectMapper(new ObjectMapper());

        ResponseEntity responseEntity = this.sut.recordChanged();
        Assert.assertEquals(200, responseEntity.getStatusCode().value());
        Assert.assertEquals("message acknowledged by client", responseEntity.getBody().toString());
    }

    @Test
    public void should_return400_whenSendOutRequestButNoSuccessCodeReturned_hmac() throws Exception {
        response.setResponseCode(500);

        when(this.signatureService.getSignedSignature(any(), any())).thenReturn(SIGNED_SIGNATURE);
        when(this.httpClient.send(any())).thenReturn(response);
        when(this.subscriptionCacheFactory.get(any())).thenReturn(getHmacSubscription());
        sut.setObjectMapper(new ObjectMapper());

        ResponseEntity responseEntity = this.sut.recordChanged();
        Assert.assertEquals(400, responseEntity.getStatusCode().value());
        Assert.assertEquals("message not acknowledged by client", responseEntity.getBody().toString());
    }

    @Test
    public void should_return400_whenSendOutRequestButNoSuccessCodeReturned_gsa() throws Exception {
        response.setResponseCode(500);

        when(this.googleIdTokenProducer.getIdToken(any(), any())).thenReturn(GOOGLE_ID_TOKEN);
        when(this.httpClient.send(any())).thenReturn(response);
        when(this.subscriptionCacheFactory.get(any())).thenReturn(getGsaSubscription());
        sut.setObjectMapper(new ObjectMapper());

        ResponseEntity responseEntity = this.sut.recordChanged();
        Assert.assertEquals(400, responseEntity.getStatusCode().value());
        Assert.assertEquals("message not acknowledged by client", responseEntity.getBody().toString());
    }

    @Test(expected = AppException.class)
    public void should_throwException_whenSubscriptionNotFound() throws Exception {
        when(this.subscriptionCacheFactory.get(any())).thenReturn(null);
        when(this.credentialHeadersProvider.getObject()).thenReturn(new DpsHeaders());
        ISubscriptionService subscriptionService = mock(SubscriptionService.class);
        when(this.subscriptionFactory.create(any())).thenReturn(subscriptionService);

        HttpResponse response = new HttpResponse();
        when(subscriptionService.query(any())).thenThrow(new SubscriptionException("error", response));
        this.sut.recordChanged();
        fail("should throw AppException");
    }

    @Test(expected = AppException.class)
    public void should_throwException_whenSubscriptionParsingErrorOccurs() throws Exception {

        when(this.subscriptionCacheFactory.get(any())).thenReturn(null);
        when(this.credentialHeadersProvider.getObject()).thenReturn(new DpsHeaders());
        ISubscriptionService subscriptionService = mock(SubscriptionService.class);
        when(this.subscriptionFactory.create(any())).thenReturn(subscriptionService);

        String jsonSubscription = this.getHmacSubscription();
        ObjectMapper objectMapper = new ObjectMapper();
        Subscription subscription = objectMapper.readValue(jsonSubscription, Subscription.class);
        List<Subscription> queryResult = new ArrayList<>();
        queryResult.add(subscription);

        when(subscriptionService.query(any())).thenReturn(queryResult);

        sut.setObjectMapper(this.objectMapper);
        when(this.objectMapper.readValue(anyString(), any(Class.class))).thenThrow(new IOException());
        this.sut.recordChanged();
        fail("should throw AppException");
    }

    @Test
    public void should_return200_whenSubscriptionGotFromRegistration() throws Exception {
        response.setResponseCode(200);

        when(this.googleIdTokenProducer.getIdToken(any(), any())).thenReturn(GOOGLE_ID_TOKEN);
        when(this.httpClient.send(any())).thenReturn(response);

        when(this.subscriptionCacheFactory.get(any())).thenReturn(null);
        when(this.credentialHeadersProvider.getObject()).thenReturn(new DpsHeaders());
        ISubscriptionService subscriptionService = mock(SubscriptionService.class);
        when(this.subscriptionFactory.create(any())).thenReturn(subscriptionService);
        sut.setObjectMapper(new ObjectMapper());

        String jsonSubscription = this.getHmacSubscription();
        ObjectMapper objectMapper = new ObjectMapper();
        // unit test purpose, not sure if this is osdu compliant
        Subscription subscription = objectMapper.readValue(jsonSubscription, Subscription.class);
        List<Subscription> queryResult = new ArrayList<>();
        queryResult.add(subscription);
        when(subscriptionService.query(any())).thenReturn(queryResult);
        ResponseEntity responseEntity = this.sut.recordChanged();
        Assert.assertEquals(200, responseEntity.getStatusCode().value());
        Assert.assertEquals("message acknowledged by client", responseEntity.getBody().toString());
    }

    @Test(expected = IllegalArgumentException.class)
    public void should_throwException_whenErrorGeneratingSignature_hmac() throws Exception {
        IllegalArgumentException ex = new IllegalArgumentException("Error generating signed signature");

        when(this.subscriptionCacheFactory.get(any())).thenReturn(getHmacSubscription());
        when(this.signatureService.getSignedSignature(any(), any())).thenThrow(ex);
        sut.setObjectMapper(new ObjectMapper());

        this.sut.recordChanged();
    }

    private String getHmacSubscription() {
        return "{\n" +
                "\t\"name\": \"testSubscription\",\n" +
                "\t\"description\": \"Description\",\n" +
                "\t\"topic\": \"records-changed\",\n" +
                "\t\"pushEndpoint\": \"http://challenge\",\n" +
                "\t\"notificationId\": \"notificationId\",\n" +
                "\t\"id\": \"id_1\",\n" +
                "\t\"createdBy\": \"test@test.com\",\n" +
                "\t\"secret\": {\n" +
                "\t\t\"secretType\": \"HMAC\",\n" +
                "\t\t\"value\": \"testsecret\"\n" +
                "\t}\n" +
                "}";
    }

    private String getGsaSubscription() {
        return "{\n" +
                "\t\"name\": \"testSubscription\",\n" +
                "\t\"description\": \"Description\",\n" +
                "\t\"topic\": \"records-changed\",\n" +
                "\t\"pushEndpoint\": \"http:///gsa-challenge\",\n" +
                "\t\"notificationId\": \"notificationId\",\n" +
                "\t\"id\": \"id_1\",\n" +
                "\t\"createdBy\": \"test@test.com\",\n" +
                "\t\"secret\": {\n" +
                "\t\t\"secretType\": \"GSA\",\n" +
                "\t\t\"value\": {\n" +
                "\t\t\t\"audience\":\"audience\",\n" +
                "\t\t\t\"key\":\"keyFile\"\n" +
                "\t\t}\n" +
                "\t}\n" +
                "}";
    }
}
